import sys
from PyQt5.QtWidgets import QApplication, QWidget
from PyQt5.QtGui import QIcon, QPixmap, QImage
from PyQt5.uic import loadUi
import opr
import comm_ard
import random
import pickle
import os
from openvino.inference_engine import IENetwork, IECore
import numpy as np
import math

import cv2


class SingleOutputPostprocessor:
    def __init__(self, output_layer):
        self.output_layer = output_layer

    def __call__(self, outputs):
        return outputs[self.output_layer].buffer[0][0]


class MultipleOutputPostprocessor:
    def __init__(self, bboxes_layer='bboxes', scores_layer='scores', labels_layer='labels'):
        self.bboxes_layer = bboxes_layer
        self.scores_layer = scores_layer
        self.labels_layer = labels_layer

    def __call__(self, outputs):
        bboxes = outputs[self.bboxes_layer].buffer[0]
        scores = outputs[self.scores_layer].buffer[0]
        labels = outputs[self.labels_layer].buffer[0]
        return [[0, label, score, *bbox] for label, score, bbox in zip(labels, scores, bboxes)]


def get_output_postprocessor(net, bboxes='bboxes', labels='labels', scores='scores'):
    if len(net.outputs) == 1:
        output_blob = next(iter(net.outputs))
        return SingleOutputPostprocessor(output_blob)
    elif len(net.outputs) >= 3:
        def find_layer(name, all_outputs):
            suitable_layers = [layer_name for layer_name in all_outputs if name in layer_name]
            if not suitable_layers:
                raise ValueError('Suitable layer for "{}" output is not found'.format(name))

            if len(suitable_layers) > 1:
                raise ValueError('More than 1 layer matched to "{}" output'.format(name))

            return suitable_layers[0]

        labels_out = find_layer(labels, net.outputs)
        scores_out = find_layer(scores, net.outputs)
        bboxes_out = find_layer(bboxes, net.outputs)

        return MultipleOutputPostprocessor(bboxes_out, scores_out, labels_out)

    raise RuntimeError("Unsupported model outputs")

class Network:
    '''
    Load and store information for working with the Inference Engine,
    and any loaded models.
    '''

    def __init__(self):
        self.plugin = None
        self.input_blob = None
        self.exec_network = None

        self.output_postprocessor = None

    def load_model(self, model, device="CPU", cpu_extension=None):
        '''
        Load the model given IR files.
        Defaults to CPU as device for use in the workspace.
        Synchronous requests made within.
        '''
        model_xml = model
        model_bin = os.path.splitext(model_xml)[0] + ".bin"

        # Initialize the plugin
        self.plugin = IECore()

        # Add a CPU extension, if applicable
        if cpu_extension and "CPU" in device:
            self.plugin.add_extension(cpu_extension, device)

        # Read the IR as a IENetwork
        # network = IENetwork(model=model_xml, weights=model_bin)
        network = self.plugin.read_network(model=model_xml, weights=model_bin)

        self.output_postprocessor = get_output_postprocessor(network)

        # Load the IENetwork into the plugin
        self.exec_network = self.plugin.load_network(network, device)

        # Get the input layer
        self.input_blob = next(iter(network.inputs))

        # Return the input shape (to determine preprocessing)
        return network.inputs[self.input_blob].shape


    def sync_inference(self, image):
        '''
        Makes a synchronous inference request, given an input image.
        '''
        self.exec_network.infer({self.input_blob: image})
        return


    def extract_output(self):
        '''
        Returns a list of the results for the output layer of the network.
        '''
        return self.output_postprocessor(self.exec_network.requests[0].output_blobs)

#进行参数的一个设定对于初始的角度包括
class App(QWidget):

    def __init__(self):
        super().__init__()

        self.ui = loadUi('tracking.ui', self)    #Load .ui file that define the window layout(buttons, tick boxoex...)

        self.setMouseTracking(True)    #allow mouse tracking(use during manual mode)
        self.manual_mode = False       #set manual mode to false to start in tracking face mode
        self.LED_ON = True             #set LED on mode
        self.CameraID = 0              #define camera ID, first camera = ID 0, second ID 1 and so on

        self.rec = True                #allows to start camera recording
        self.cap = cv2.VideoCapture(self.CameraID)  #define open CV camera recording
        self.cap.set(3, 960)                        #set capture width
        self.cap.set(4, 540)                        #set capture height
        #The bigger the image is, the longer the processing is goint to take to process it
        ## My computer is a bit shit so I kept it quite small .

        self.min_tilt = 22          #minimum tilt angle in degree (up/down angle)需要对自己的摄像头进行一个矫正上下的角度
        self.max_tilt = 80          #maximum tilt angle in degree
        self.current_tilt = 0       #current tilt (info received from arduino and displayed in LCD numbers)
        self.target_tilt = 45      #the tilt angle you need to reach

        self.min_pan = 20            #minimum pan angle in degree(left/ right angle)左右的角度
        self.max_pan = 100         #maximum pan angle in degree
        self.current_pan = 80        #current pan (info received from arduino and displayed in LCD numbers)
        self.target_pan = 90       #the pan angle you need to reach

        self.roam_target_pan = 90
        self.roam_target_tilt = 90
        self.roam_pause = 40      #amount of frame the camera is going to pause for when roam tilt or pan target reached
        self.roam_pause_count = self.roam_pause   #current pause frame count


        self.is_connected = False    #boolean defining if arduino is connected

        self.InvertPan = False       #allows pan to be inverted
        self.InvertTilt = False      #allows tilt to be inverted

        self.face_detected = False     #define if a face is detected
        self.target_locked = False     #define if detected face is close enough to the center of the captured image
        self.max_target_distance = 40  #minimum distance between face/center of image for setting target locked
        self.max_empty_frame = 50      #number of empty frame (no face detected) detected before starting roaming
        self.empty_frame_number = self.max_empty_frame   #current empty frame count

        self.ard = comm_ard.ard_connect(self)     #create object allowing communicationn with arduino进行arduino的连接通讯


        '''爱德华start'''
        self.inference_network = Network()
        self.n, self.c, self.h, self.w = self.inference_network.load_model(
            'D:/mdeldownloader/face-detection-adas-0001/FP32/face-detection-adas-0001.xml', 'CPU', None)
        self.prob_threshold = 0.5
        '''爱德华end'''

        self.initUI()    #set up UI( see below )

    def initUI(self):     #UI related stuff
        self.setWindowTitle('LFK Tracker')              #set window title
        self.label = self.ui.label                      #set label (it will be used to display the captured images)
        self.QuitButton = self.ui.QuitButton            #set quit button
        self.PauseButton = self.ui.PauseButton          #set pause button
        self.Pan_LCD = self.ui.Pan_LCD                  #and so on...
        self.Tilt_LCD = self.ui.Tilt_LCD                #...
        self.Manual_checkbox = self.ui.Manual_checkbox  #...
        self.ConnectButton = self.ui.ConnectButton
        self.COMlineEdit = self.ui.COMlineEdit
        self.COMConnectLabel = self.ui.COMConnectLabel
        self.UpdateButton = self.ui.UpdateButton
        self.MinTiltlineEdit = self.ui.MinTiltlineEdit
        self.MaxTiltlineEdit = self.ui.MaxTiltlineEdit
        self.InvertTilt_checkbox = self.ui.InvertTilt_checkbox
        self.InvertTilt = self.InvertTilt_checkbox.isChecked()
        self.MinPanlineEdit = self.ui.MinPanlineEdit
        self.MaxPanlineEdit = self.ui.MaxPanlineEdit
        self.InvertPan_checkbox = self.ui.InvertPan_checkbox
        self.InvertPan = self.InvertPan_checkbox.isChecked()
        self.TiltSensivityEdit = self.ui.TiltSensivityEdit
        self.TiltSensivity = 1
        self.PanSensivityEdit = self.ui.PanSensivityEdit
        self.PanSensivity = 1
        self.LED_checkbox = self.ui.LED_checkbox
        self.CameraIDEdit = self.ui.CameraIDEdit

        self.QuitButton.clicked.connect(self.quit)               #bind quit button to quit method
        self.PauseButton.clicked.connect(self.toggle_recording)  #bind pause button to pause method
        self.Manual_checkbox.stateChanged.connect(self.set_manual_mode)  #bla bla bla
        self.ConnectButton.clicked.connect(self.connect)
        self.UpdateButton.clicked.connect(self.update_angles)

        self.load_init_file()
        self.update_angles() #update angle method

        self.record()  #start recording

    def load_init_file(self):#具体功能不了解，只是简单了解功能
        #this method will allow to reload the latest values entered in text boxes even after closing the software
        try:         #try to open init file if existing
            with open('init.pkl', 'rb') as init_file:
                var = pickle.load(init_file)  #load all variable and update text boxes
                self.COMlineEdit.setText(var[0])
                if(var[4]):
                    self.MinTiltlineEdit.setText(str(var[2]))
                    self.MaxTiltlineEdit.setText(str(var[1]))
                else:
                    self.MinTiltlineEdit.setText(str(var[1]))
                    self.MaxTiltlineEdit.setText(str(var[2]))
                self.TiltSensivityEdit.setText(str(var[3]))
                self.InvertTilt_checkbox.setChecked(var[4])
                if (var[8]):
                    self.MinPanlineEdit.setText(str(var[6]))
                    self.MaxPanlineEdit.setText(str(var[5]))
                else:
                    self.MinPanlineEdit.setText(str(var[5]))
                    self.MaxPanlineEdit.setText(str(var[6]))
                self.PanSensivityEdit.setText(str(var[7]))
                self.InvertPan_checkbox.setChecked(var[8])
                #self.CameraIDEdit.setText(str(var[9]))
                self.LED_checkbox.setChecked(var[10])
            print(var)
            #set variables
        except:
            pass

    def save_init_file(self):
        init_settings = [self.COMlineEdit.text(),
        self.min_tilt, self.max_tilt, self.TiltSensivity, self.InvertTilt,
        self.min_pan, self.max_pan, self.PanSensivity, self.InvertPan,
        self.CameraID, self.LED_ON]
        with open('init.pkl', 'wb') as init_file:
            pickle.dump(init_settings, init_file)


    def connect(self):    #set COM port from text box if arduino not already connected
        if(not self.is_connected):
            port = self.COMlineEdit.text()
            if (self.ard.connect(port)):    #set port label message
                self.COMConnectLabel.setText("..................... Connected to port : " + port + " ......................")
            else:
                self.COMConnectLabel.setText(".................... Cant connect to port : " + port + " .....................")

    def update_angles(self):  #update variables from text boxes
        try:
            self.InvertTilt = self.InvertTilt_checkbox.isChecked()
            self.InvertPan = self.InvertPan_checkbox.isChecked()
            self.TiltSensivity = float(self.TiltSensivityEdit.text())
            self.PanSensivity = float(self.PanSensivityEdit.text())
            self.LED_ON = self.LED_checkbox.isChecked()

            self.cap.release()    #camera need to be released in order to update the camera ID (if changed)
            self.CameraID = int(self.CameraIDEdit.text())
            self.cap = cv2.VideoCapture(self.CameraID)

            if(self.InvertPan):
                self.max_pan = int(self.MinPanlineEdit.text())
                self.min_pan = int(self.MaxPanlineEdit.text())
            else:
                self.min_pan = int(self.MinPanlineEdit.text())
                self.max_pan = int(self.MaxPanlineEdit.text())

            if(self.InvertTilt):
                self.max_tilt = int(self.MinTiltlineEdit.text())
                self.min_tilt = int(self.MaxTiltlineEdit.text())
            else:
                self.min_tilt = int(self.MinTiltlineEdit.text())
                self.max_tilt = int(self.MaxTiltlineEdit.text())

            self.save_init_file()
            print("values updated")
        except:
            print("can't update values")

    def mouseMoveEvent(self, event):#这段是进行一个跟随鼠标的位置来进行的移动

        #the position of the mouse is tracked avec the window and converted to a pan and tilt amount
        #for example if mouse completely to the left-> pan_target = 0(or whatever minimum pan_target value is)
        # if completely to the right-> pan_target = 180(or whatever maximum pan_target value is)
        #same principal for tilt

        if(self.manual_mode): #if manual mode selected好像这部分是关键，应该是存在一个映射关系opr。remap
            if(35<event.y()<470 and 70<event.x()<910):
                if(self.InvertTilt):  #if invert tilt selected values will be mapped in the opposite direction
                    self.target_tilt = opr.remap(event.y(), self.max_tilt, self.min_tilt, 470, 35)
                    #(470,35 allows the mouse to be tracked only over the image. yes...it is very lazy)
                else:
                    self.target_tilt =  opr.remap(event.y(), self.min_tilt, self.max_tilt, 35, 470)

                if (self.InvertPan):
                    self.target_pan = opr.remap(event.x(), self.max_pan, self.min_pan, 910, 70)  # event.x()
                else:
                    self.target_pan = opr.remap(event.x(), self.min_pan, self.max_pan, 70, 910)  # event.x()

    def update_LCD_display(self):

        #update servo angle values sent by the arduino
        ##actually pretty useless since the arduino return its target value and not its actual real world position
        self.Pan_LCD.display(self.current_pan)
        self.Tilt_LCD.display(self.current_tilt)

    def quit(self):
        print('Quit')
        self.rec = False
        sys.exit()

    def closeEvent(self, event):
        self.quit()# call quit method when cross pressed

    def set_manual_mode(self):
        self.manual_mode = self.Manual_checkbox.isChecked()
        if(not self.manual_mode):          #if not in manual mode
            self.random_servos_position()  #select a random pan and tilt target
        print(self.manual_mode)

    def random_servos_position(self):
        self.target_tilt = random.uniform(self.min_tilt, self.max_tilt)
        self.target_pan = random.uniform(self.min_pan, self.max_pan)


    def toggle_recording(self):
        if(self.rec):
            self.rec = False                   #stop recording
            self.PauseButton.setText("Resume") #change pause button text
        else:
            self.rec = True
            self.PauseButton.setText("Pause")
            self.record()

    def preprocessing(self, input_image):
        '''
        Given an input image, height and width:
        - Resize to width and height
        - Transpose the final "channel" dimension to be first
        - Reshape the image to add a "batch" of 1 at the start
        '''
        in_frame = cv2.resize(input_image, (self.w, self.h))
        in_frame = in_frame.transpose((2, 0, 1))  # Change data layout from HWC to CHW
        in_frame = in_frame.reshape((self.n, self.c, self.h, self.w))

        return in_frame

    def record(self):  #video recording

        while(self.rec):
            ret, img = self.cap.read() #CAPTURE IMAGE

            if(self.is_connected):            #if arduino connected
                if(self.manual_mode):         #and manual mode on
                    processed_img = img       #don't process image
                else:
                    processed_img = self.image_process(img) #process image (check for faces and draw circle and cross)
            else:                             #if arduino  not connected
                processed_img = img           #don't process image

            self.update_GUI(processed_img)    #update image in window
            cv2.waitKey(0)                    #no delay between frames

            self.move_servos() #move servos

            if (not self.rec):     #allows while loop to stop if pause button pressed
                break

    def update_GUI(self, openCV_img): # Convert openCV image and update pyQt label

        try: #if this doesn't work check camera ID
            openCV_img = cv2.resize(openCV_img, (960, 540))  #this is stretching the image a bit but if not done it won't fit the UI
            height, width, channel = openCV_img.shape
            bytesPerLine = 3 * width
            qImg = QImage(openCV_img.data, width, height, bytesPerLine, QImage.Format_RGB888).rgbSwapped()#qt编程相关

            pixmap = QPixmap(qImg)
            self.label.setPixmap(pixmap)

        except:
            self.label.setText("check camera ID")

        self.show()

    def move_servos(self):
        if (self.is_connected):

            if self.LED_ON and not self.manual_mode:
                if not self.face_detected: #set led mode (0:red, 1:yellow 2:green)
                    led_mode = 0
                else:
                    if self.target_locked:
                        led_mode = 1
                    else :
                        led_mode = 2

            elif self.LED_ON and self.manual_mode:
                led_mode = 3 #turn all led's on
            else:
                led_mode = 4 #turn led's off

            data_to_send = "<" + str(int(self.target_pan)) + "," + str(int(self.target_tilt)) + "," + str(led_mode) + ">"
            print(data_to_send)#就是屏幕上显示出来的pan，tilt的值
            self.ard.runTest(data_to_send)
            #the data sent to the arduino will look something like the this (<154, 23, 0>)
            #the arduino will look for the start character "<"
            #then save everything following until it finds the end character ">"
            #at that point the arduino will have saved a message looking like this "154, 23, 0"
            #it will then split the message at every coma and use the pieces of data to move the servos acordingly

            #And yes I should have used bytes or something a bit better
            #I wrote this a while back and at that time I had even more to learn than I do now!
            #I still lazy and can't be asked to change it now as it works .


    def roam(self):
        if(self.roam_pause_count < 0 ):      #if roam count inferior to 0

            self.roam_pause_count = self.roam_pause                                        #reset roam count
            self.roam_target_pan = int(random.uniform(self.min_pan, self.max_pan))
            self.roam_target_tilt = int(random.uniform(self.min_tilt, self.max_tilt))

        else:        #if roam count > 1
                     #increment pan target toward roam target
            if (int(self.target_pan) > self.roam_target_pan):
                self.target_pan -= 1
            elif (int(self.target_pan) < self.roam_target_pan):
                self.target_pan += 1
            else:    #if roam target reached decrease roam pause count
                self.roam_pause_count -= 1

            if (int(self.target_tilt) > self.roam_target_tilt):
                self.target_tilt -= 1
            elif (int(self.target_tilt) < self.roam_target_tilt):
                self.target_tilt += 1
            else:
                self.roam_pause_count -= 1


    def image_process(self, img):  #handle the image processing
        #to add later : introduce frame scipping (check only 1 every nframe)

        '''爱德华start'''
        frame_h, frame_w = img.shape[:2]
        preprocessed_image = self.preprocessing(img)
        # Perform synchronous inference on the image
        self.inference_network.sync_inference(preprocessed_image)
        # Obtain the output of the inference request
        output = self.inference_network.extract_output()#

        for obj in output:
            # Draw only objects when probability more than specified threshold
            if obj[2] > self.prob_threshold:
                xmin = int(obj[3] * frame_w)
                ymin = int(obj[4] * frame_h)
                xmax = int(obj[5] * frame_w)
                ymax = int(obj[6] * frame_h)
                class_id = int(obj[1])
                # Draw box and label\class_id
                color = (min(class_id * 12.5, 255), min(class_id * 7, 255), min(class_id * 5, 255))
                cv2.rectangle(img, (xmin, ymin), (xmax, ymax), color, 2)

                center_face_X = int((xmin + xmax)/ 2)
                center_face_Y = int((ymin + ymax)/ 2)
                height, width, channels = img.shape
                distance_from_center_X = (center_face_X - width / 2) / 220  # why? can't remember why I did this
                distance_from_center_Y = (center_face_Y - height / 2) / 195  # why?

                target_distance = math.sqrt((distance_from_center_X * 220) ** 2 + (
                            distance_from_center_Y * 195) ** 2)  # 进行距离的计算，勾股定理calculate distance between image center and face center
                if target_distance < self.max_target_distance:  # set added geometry colour
                    locked = True
                    color = (0, 255, 0)
                else:
                    locked = False
                    color = (0, 0, 255)#BGR

                cv2.rectangle(img, (center_face_X - 10, center_face_Y), (center_face_X + 10, center_face_Y),
                              # draw first line of the cross
                              color, 2)
                cv2.rectangle(img, (center_face_X, center_face_Y - 10), (center_face_X, center_face_Y + 10),
                              # draw second line of the cross
                              color, 2)

                cv2.circle(img, (int(width / 2), int(height / 2)), int(self.max_target_distance), color,
                           2)  # draw circle

                processed_img = [True, img, distance_from_center_X, distance_from_center_Y, locked]#processed_img[0][1][2][3][4]
                break
            else:
                processed_img = [False]

        '''爱德华end'''


        #processed_img = opr.find_face(img, self.max_target_distance)  # try to find face and return processed image
        # if face found during processing , the data return will be as following :
        #[True, image_to_check, distance_from_center_X, distance_from_center_Y, locked]
        #if not it will just retun False

        if(processed_img[0]):             #if face found
            self.face_detected = True
            self.empty_frame_number = self.max_empty_frame  #reset empty frame count
            self.target_locked = processed_img[4]
            self.calculate_camera_move(processed_img[2], processed_img[3])  # calculate new targets depending on distance between face and image center
            return processed_img[1]
        else:
            self.face_detected = False
            self.target_locked = False
            if(self.empty_frame_number> 0):
                self.empty_frame_number -= 1  #decrease frame count until it equal 0
            else:
                self.roam()              #then roam就是进行一个巡视
            return img


    def calculate_camera_move(self, distance_X, distance_Y):
        #print(distance_X,distance_Y)
        # distance_Y=distance_Y*(-1)
        # distance_X=distance_X*(-1)
        #self.target_pan += distance_X * self.PanSensivity

        if(self.InvertPan): #handle inverted pan
            self.target_pan -= distance_X * self.PanSensivity#distance——x  target_pan=90`
            if(self.target_pan>self.min_pan):
                self.target_pan = self.min_pan
            elif (self.target_pan < self.max_pan):
                self.target_pan = self.max_pan

        else:
            self.target_pan += distance_X * self.PanSensivity#居然是距离乘以灵敏度
            #print(self.target_pan)
            if(self.target_pan>self.max_pan):
                self.target_pan = self.max_pan
            elif (self.target_pan < self.min_pan):
                self.target_pan = self.min_pan


        #self.target_tilt += distance_Y * self.TiltSensivity

        if(self.InvertTilt): #handle inverted tilt
            self.target_tilt -= distance_Y * self.TiltSensivity
            if(self.target_tilt>self.min_tilt):
                self.target_tilt = self.min_tilt
            elif (self.target_tilt < self.max_tilt):
                self.target_tilt = self.max_tilt
        else:
            self.target_tilt += distance_Y * self.TiltSensivity
            if(self.target_tilt>self.max_tilt):
                self.target_tilt = self.max_tilt
            elif (self.target_tilt < self.min_tilt):
                self.target_tilt = self.min_tilt

if __name__ == '__main__':
    app = QApplication(sys.argv)
    ex = App()
    app.exec_()
    #sys.exit(app.exec_())
