package com.example.wanghu;

import androidx.appcompat.app.AppCompatActivity;
import android.hardware.Sensor;
import android.hardware.SensorEvent;
import android.hardware.SensorEventListener;
import android.hardware.SensorManager;
import android.os.Bundle;
import android.os.Handler;
import android.os.StrictMode;
import android.view.View;
import android.widget.TextView;
import java.io.IOException;
import java.io.OutputStream;
import java.net.Socket;
import java.util.Date;

public class MainActivity extends AppCompatActivity implements SensorEventListener {

    private SensorManager sensorManager;
    private Sensor linear;
    private Sensor rotation;
    private Sensor gravity;
    private float[] linearArray= new float[3];
    private float[] rotationArray= new float[4];
    private float[] gravityArray = new float[3];

    private Handler handler;
    private Runnable runnable;

    private int counter = 0; // 记录数据条目
    private int still_counter = 0; // 记录数据条目
    private boolean isStart = false; // 记录是否开始记录数据

    private int steps = 0; // 记录步数
    private float distance = 0; // 记录距离
    private float angle = 0; // 记录角度

    private Socket mSocket;
    private OutputStream mOutputStream;

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_main);

        sensorManager = (SensorManager) getSystemService(SENSOR_SERVICE);
        linear = sensorManager.getDefaultSensor(Sensor.TYPE_LINEAR_ACCELERATION);
        rotation = sensorManager.getDefaultSensor(Sensor.TYPE_ROTATION_VECTOR);
        gravity = sensorManager.getDefaultSensor(Sensor.TYPE_GRAVITY);

        StrictMode.setThreadPolicy(new StrictMode.ThreadPolicy.Builder()
                .detectDiskReads().detectDiskWrites().detectNetwork()
                .penaltyLog().build());
        StrictMode.setVmPolicy(new StrictMode.VmPolicy.Builder()
                .detectLeakedSqlLiteObjects().penaltyLog().penaltyDeath()
                .build());

        new Thread(new Runnable() {
            @Override
            public void run() {
                while (true) {
                    try {
                        String sendingData = String.valueOf(steps)
                                + ',' + distance
                                + ',' + angle;
                        mSocket = new Socket("192.168.31.221",8999);
                        if(mOutputStream != null) {
                            mOutputStream.write(sendingData.getBytes());
                            mOutputStream.flush();
                        }
                        mOutputStream = mSocket.getOutputStream();
                        LogUtil.d("success", "socket连接成功");

                        Thread.sleep(50);
                    } catch (IOException | InterruptedException e) {
                        e.printStackTrace();
                        LogUtil.d("failed", "socket连接失败");
                    }
                }
            }
        }).start();
    }

    @Override
    protected void onDestroy() {
        super.onDestroy();
    }

    /** Called when the user touches the "开始" button */
    public void collectData(View view) {
        if(isStart) {
            return;
        }

        Date date = new Date();
        LogUtil.i("运行开始时间", String.valueOf(date));

        isStart = true;
        final TextView textView_status_label = (TextView)findViewById(R.id.status_label);
        final TextView textView_total_steps = (TextView)findViewById(R.id.total_steps);
        final TextView textView_total_distance = (TextView)findViewById(R.id.total_distance);
        final TextView textView_angle_view = (TextView)findViewById(R.id.angle_view);

        /** 步伐检测相关参数 **/
        final double frequency = 92;
        final double offset = frequency/100; // 考虑数据采集频率偏差，针对于frequency不为100的情况
        double g = 9.794;
        final int slide = (int) Math.ceil(40 * offset); // 滑动窗口（100Hz的采样数据），窗口需为整数
        // 行人加速度阈值
        final double min_acceleration = 0.2 * g; // 0.2g
        final double max_acceleration = 2 * g;   // 2g
        // 峰值间隔(s)
        final double min_interval = 0.4; // max_interval = 1
        final double[] temp_step = {0, 0, 0}; // [index, acceleration, angle]
        final double[] last_step = {0, 0, 0}; // [index, acceleration, angle]

        handler = new Handler();
        runnable = new Runnable() {
            @Override
            public void run() {
                ++counter;
                if(counter%50 == 0) {
                    textView_status_label.setText("软件运行中：已处理" + counter + "条数据");
                    textView_total_steps.setText("总步数：" + steps);
                    textView_total_distance.setText("总距离："+ distance);
                    textView_angle_view.setText("航向角：" + angle);
                }

                /** 步伐检测 **/
                double a_vertical = coordinate_conversion(linearArray, gravityArray);
                angle = quaternion2euler(rotationArray)[2];

                // 条件1：峰值在0.2g~2g之间，寻找一个窗口中的最大值
                if(a_vertical>=temp_step[1] && a_vertical>=min_acceleration && a_vertical<=max_acceleration) {
                    still_counter = 0;
                    temp_step[0] = counter;
                    temp_step[1] = a_vertical;
                    temp_step[2] = angle;
                } else {
                    still_counter += 1;
                    if(still_counter >= frequency*3) {
                        still_counter = 0;
                        steps = 0;
                        distance = 0;
                    }
                }

                // 以slide为滑动窗检测峰值
                if(counter>0 && counter%slide ==0) {
                    // 记录第一步
                    if(last_step[0] == 0 && temp_step[0] != 0) {
                        last_step[0] = temp_step[0];
                        last_step[1] = temp_step[1];
                        last_step[2] = temp_step[2];
                        steps += 1;
                        distance += step_stride(last_step[1]);
                    } else { // 从第二步开始与上一步进行比较
                        // 条件2：两个峰值之前间隔至少大于0.4s*frequency
                        if((temp_step[0] - last_step[0]) < min_interval*frequency) {
                            if(temp_step[1] > last_step[1]) {
                                distance -= step_stride(last_step[1]);
                                last_step[0] = temp_step[0];
                                last_step[1] = temp_step[1];
                                last_step[2] = temp_step[2];
                                distance += step_stride(last_step[1]);
                            }
                        } else {
                            last_step[0] = temp_step[0];
                            last_step[1] = temp_step[1];
                            last_step[2] = temp_step[2];
                            steps += 1;
                            distance += step_stride(last_step[1]);
                        }
                    }

                    temp_step[0] = 0;
                    temp_step[1] = 0;
                    temp_step[2] = 0;
                }

                handler.postDelayed(this, 10); // 100Hz
            }
        };

        handler.postDelayed(runnable, 0);
    }

    // 四元数转化为欧拉角
    public float[] quaternion2euler(float[] rotation) {
        float x = rotation[0];
        float y = rotation[1];
        float z = rotation[2];
        float w = rotation[3];
        float pitch = (float)Math.asin(2*(w*y-z*x));
        float roll = (float)Math.atan2(2*(w*x+y*z),1-2*(x*x+y*y));
        float yaw = (float)Math.atan2(2*(w*z+x*y),1-2*(z*z+y*y));
        float[] result = {pitch, roll, yaw};
        return result;
    }

    /** 坐标转换，获取垂直向加速度 **/
    public double coordinate_conversion(float[] linearArray,  float[] gravityArray) {
        float g_y = gravityArray[1];
        float g_z = gravityArray[2];
        float linear_y = linearArray[1];
        float linear_z = linearArray[2];

        float theta = (float) Math.atan(Math.abs(g_z/g_y));
        float a_vertical = (float)(linear_y*Math.cos(theta) + linear_z*Math.sin(theta));

        return a_vertical;
    }

    public float step_stride(double max_acceleration) {
        float distance = (float)Math.pow(max_acceleration, 1.0/4) / 2;
        return distance;
    }

    /** Called when the user touches the "结束" button **/
    public void stopProcess(View view) throws IOException {
        if(!isStart) {
            return;
        }

        Date date = new Date();
        LogUtil.i("运行结束时间", String.valueOf(date));

        isStart = false;
        handler.removeCallbacks(runnable);
        TextView textView = (TextView)findViewById(R.id.status_label);
        textView.setText("已停止操作。");

        LogUtil.i("操作数据数目", String.valueOf(counter));

        steps = 0;
        distance = 0;
        counter = 0;
        mSocket = null;
        mOutputStream = null;
    }

    @Override
    public final void onAccuracyChanged(Sensor sensor, int accuracy) {
        // Do something here if sensor accuracy changes.
    }

    @Override
    /** 监听传感器数据的变化，同时讲数据更新到自定义的数组变量中 **/
    public final void onSensorChanged(SensorEvent event) {
        if (event.sensor.getType() == Sensor.TYPE_LINEAR_ACCELERATION) {
            // 获取加速度传感器的三个参数
            linearArray[0] = event.values[0];
            linearArray[1] = event.values[1];
            linearArray[2] = event.values[2];
        }
        if (event.sensor.getType() == Sensor.TYPE_ROTATION_VECTOR) {
            rotationArray[0] = event.values[0];
            rotationArray[1] = event.values[1];
            rotationArray[2] = event.values[2];
            rotationArray[3] = event.values[3];
        }
        if (event.sensor.getType() == Sensor.TYPE_GRAVITY) {
            // 获取加速度传感器的三个参数
            gravityArray[0] = event.values[0];
            gravityArray[1] = event.values[1];
            gravityArray[2] = event.values[2];
        }
    }

    @Override
    protected void onResume() {
        super.onResume();
        // 注册传感器监听函数
        sensorManager.registerListener(this, linear, SensorManager.SENSOR_DELAY_NORMAL);
        sensorManager.registerListener(this, rotation, SensorManager.SENSOR_DELAY_NORMAL);
        sensorManager.registerListener(this,gravity, SensorManager.SENSOR_DELAY_NORMAL);
    }

    @Override
    protected void onPause() {
        super.onPause();
        // 注销监听函数
        sensorManager.unregisterListener(this);
    }
}
